"""dash_charts."""

from loguru import logger

__version__ = '0.1.0'
__pkg_name__ = 'dash_charts'

logger.disable(__pkg_name__)

# ====== Above is the recommended code from calcipy_template and may be updated on new releases ======
