body {
    margin: 0;
    border: 0;
    height: 100%;
    min-height: 0;
}

/* Override bootstrap max-width to better fill the window */
.container {
    max-width: 95%;
}

/* Add a class specifically for sliders that need additional height */
.rc-slider-with-legend {
    min-height: 45px;
}

/* FIXME: Implement user-customizable styles
    User specifies directory with assets, then the files are copied to the Dash asset dir
    as user.js, user.css, etc. */
iframe#plaid-link-iframe-1 {
    height: 100vh;
}

/* Make sure the date range picker appears above all elements */
.DateRangePicker_picker {
    z-index: 999;
}
